#!/usr/bin/env python3

import sys
import os
import argparse
from datetime import datetime

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apcommon import iniparser
from apcommon import dockerhelper
# from apcommon import partitioning
from ApbootstrapInitialBoot import ApBootstrapInitialBoot
from ApbootstrapAutopilotIni import ApBootstrapAutopilotINI
from ApbootstrapMachineEnv import ApBootstrapMachineEnv
from ApbootstrapAptProxy import ApBootstrapAptProxy


class Bootstrap(object):
    def __init__(self,
                 apbootstrap_ini_path: str="",
                 apbootstrap_reboot_path: str="",
                 linux_bootstrap_ini_path: str="",
                 autopilot_ini_path: str="",
                 global_environment_sh_path: str="",
                 apt_proxy_file: str="",
                 apt_proxy_server: str="",
                 sections_to_reset: list=[],
                 initial_boot_csv_path: str="",
                 initial_boot_csv_destination: str="",
                 watchdog_csv_dir: str=""):

        # Variables passed in during initialization
        self.apbootstrap_ini_path = apbootstrap_ini_path
        self.apbootstrap_reboot_path = apbootstrap_reboot_path
        self.linux_bootstrap_ini_path = linux_bootstrap_ini_path
        self.autopilot_ini_path = autopilot_ini_path
        self.global_environment_sh_path = global_environment_sh_path
        self.apt_proxy_file = apt_proxy_file
        self.apt_proxy_server = apt_proxy_server
        self.sections_to_reset = sections_to_reset
        self.initial_boot_datetime = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.initial_boot_csv_path = initial_boot_csv_path
        self.initial_boot_csv_destination = initial_boot_csv_destination
        self.watchdog_csv_dir = watchdog_csv_dir

        # Initialize all objects and parsers so we have them available right away
        self.apbootstrap_ini_parser_object = iniparser.INIParser(self.apbootstrap_ini_path)
        self.dockerhelper_object = dockerhelper.DockerHelper()

        self.apbootstrap_initial_boot = ApBootstrapInitialBoot(self.initial_boot_datetime,
                                                               self.initial_boot_csv_path,
                                                               self.initial_boot_csv_destination,
                                                               self.watchdog_csv_dir)

        self.docker_bridge_ip = self.get_docker_bridge_ip()
        self.apbootstrap_autopilot_ini = ApBootstrapAutopilotINI(self.linux_bootstrap_ini_path,
                                                                 self.autopilot_ini_path)

        self.apbootstrap_machine_env = ApBootstrapMachineEnv(self.linux_bootstrap_ini_path,
                                                             self.global_environment_sh_path,
                                                             self.docker_bridge_ip)

        self.apbootstrap_apt_proxy = ApBootstrapAptProxy(self.apt_proxy_file,
                                                         self.apt_proxy_server)

        # Default sections
        self.default_sections = ["initial_boot", "autopilot_ini", "machine_env", "apt_proxy"]
        # Default options
        self.default_options = {"Execute": "True", "Success": "False", "LastRun": ""}
        # Used to set a section as Executed after it's run
        self.set_completed_options = {"Execute": "False", "Success": "True", "LastRun": datetime.now().strftime("%Y-%m-%d %H:%M:%S")}

        # Only write changes to apbootstrap.ini if the contents change
        self.changed_apbootstrap_ini = False

    # Function to get docker bridge IP
    def get_docker_bridge_ip(self) -> str:
        docker_bridge_ip = self.dockerhelper_object.get_docker_bridge()

        return docker_bridge_ip

    # Generic abstraction function to check if an ini file exists on file system
    def check_apbootstrap_ini_exists(self) -> bool:
        if (os.path.isfile(self.apbootstrap_ini_path)):
            return True
        else:
            return False

    # Generic abstraction function to create a fresh ini file in memory
    def create_apbootstrap_ini(self) -> bool:
        if (not self.apbootstrap_ini_parser_object.create()):
            return False

        return True

    # Generic abstraction function to call read on any iniparser object
    def read_apbootstrap_ini(self) -> bool:
        if (self.apbootstrap_ini_parser_object.read()):
            return True
        else:
            return False

    # Generic abstraction function to call write on any iniparser object
    def write_apbootstrap_ini(self) -> bool:
        if(not self.apbootstrap_ini_parser_object.write()):
            return False

        return True

    # Used for debug to print out all of the apbootstrap.ini file contents
    def print_all_apbootstrap_ini(self) -> bool:
        if(not self.apbootstrap_ini_parser_object.print_all()):
            return False

        return True

    # When creating a new apbootstrap.ini file in memory, pre-populate it with default values
    # It needs to know what to execute right away on first run
    def populate_new_apbootstrap_ini(self) -> bool:
        for section in self.default_sections:
            if (not self.apbootstrap_ini_parser_object.add_section(section)):
                return False

            for option, value in self.default_options.items():
                if (not self.apbootstrap_ini_parser_object.set(section, option, value)):
                    return False

        # if (not self.apbootstrap_ini_parser_object.add_section("Partitioning")):
        #    return False
        # if (not self.apbootstrap_ini_parser_object.set("Partitioning","Execute","True")):
        #    return False

        return True

    # If apbootstrap.ini needs to re-run any sections due to updates, etc.  This will instruct it to reset certain
    # sections to be run again next time it executes
    # This is done via commandline using option -r (see bottom of file)
    def reset_apbootstrap_ini_sections(self) -> bool:
        for section in self.sections_to_reset:
            # Create missing section if it doesn't exist and if it is one of the default sections that should be
            # there already
            if (not self.apbootstrap_ini_parser_object.has_section(section)) and (section in self.default_sections):
                print("apbootstrap: Default section: " + section + " does not exist (apbootstrap_ini_parser_object.has_section)")
                if (not self.apbootstrap_ini_parser_object.add_section(section)):
                    print("apbootstrap: Tried to create missing section: " + section + " but failed (apbootstrap_ini_parser_object.add_section)")
                    return False

            # Enable execution of section
            if (not self.apbootstrap_ini_parser_object.set(section, "Execute", "True")):
                return False
            # Set success to False since we need to run it again now that it's reset
            if (not self.apbootstrap_ini_parser_object.set(section, "Success", "False")):
                return False

        return True

    # Returns all sections from pre-defined sections for unit testing
    def get_apbootstrap_all_sections(self) -> list:
        return self.default_sections

    # This function is used to get all the sections and values of the apbootstrap.ini work so we can
    # verify it via unit testing
    def get_apbootstrap_all_sections_and_values(self) -> dict:
        return self.apbootstrap_ini_parser_object.get_all_sections_and_options()

    # Function that determines whether to create a new apbootstrap.ini if one doesn't exist or to use an
    # existing one if it's being changed
    def configure_apbootstrap_ini(self) -> bool:
        if (not self.check_apbootstrap_ini_exists() or not self.check_default_sections_options_exist()):
            print("apbootstrap: apbootstrap.ini does not exist or not all sections exist; must create a new one")
            if (not self.create_apbootstrap_ini()):
                print("apbootstrap: Could not create new apbootstrap ini file (apbootstrap_ini_parser_object.create_ini)")
                return False

            if (not self.populate_new_apbootstrap_ini()):
                print("apbootstrap: Could not populate new apbootstrap ini file ("
                      "apbootstrap_ini_parser_object.populate_new_apbootstrap_ini)")
                return False

            return True

        return True

    # Verify that default sections and options exist so we don't throw errors when checking their values
    def check_default_sections_options_exist(self) -> bool:
        for section in self.default_sections:
            if (not self.apbootstrap_ini_parser_object.has_section(section)):
                print("section missing: " + section)
                return False

            for option, value in self.default_options.items():
                if (not self.apbootstrap_ini_parser_object.has_option(section, option)):
                    print("section: " + section + " option missing: " + option)
                    return False

        return True

    # Sets section passed in as executed in the apbootstrap.ini file
    def set_section_executed(self, section: str) -> bool:
        if (not self.apbootstrap_ini_parser_object.remove_section(section)):
            return False

        if (not self.apbootstrap_ini_parser_object.add_section(section)):
            return False

        for option, value in self.set_completed_options.items():
            if (not self.apbootstrap_ini_parser_object.set(section, option, value)):
                return False

        self.changed_apbootstrap_ini = True

        return True

    # Checks the apbootstrap.ini file and determines whether to execute each section passed in
    def check_if_section_should_be_executed(self, section: str) -> bool:
        if(self.apbootstrap_ini_parser_object.has_section(section)):
            if (self.apbootstrap_ini_parser_object.getboolean(section, "Execute")) or not (
                    self.apbootstrap_ini_parser_object.has_option(section, "Execute")):
                print("Section: " + section + " -> EXECUTE")

                return True
            else:
                print("Section: " + section + " -> DO NOT EXECUTE")
                return False
        else:
            return False

    # Checks the apbootstrap.ini file and determines whether each section passed in succeeded in its run
    def check_if_section_execution_succeeded(self, section: str) -> bool:
        if(self.apbootstrap_ini_parser_object.has_section(section)):
            if (self.apbootstrap_ini_parser_object.getboolean(section, "Success")):
                return True
            else:
                return False
        else:
            return False

    ##########################
    # PARTITIONING GOES HERE #
    ##########################

    def main(self):
        # If we're resetting sections that need to be run again, use this path
        # If the ini file is not generated yet on initial run we do not run this first but instead create
        # the ini file and run it normally since it will already re-run everything
        # The reset is redundant in this case
        if self.sections_to_reset:
            print("Sections to reset:")
            print(self.sections_to_reset)

            if (not self.check_apbootstrap_ini_exists()):
                print("apbootstrap: Could not find apbootstrap ini file (main.check_ini_exists)")
                sys.exit(1)

            if (not self.read_apbootstrap_ini()):
                print("apbootstrap: Could not read apbootstrap ini file (main.read_apbootstrap_ini)")
                sys.exit(1)

            if (not self.check_default_sections_options_exist()):
                print("apbootstrap: Could not read apbootstrap ini file (main.check_default_sections_options_exist)")
                sys.exit(1)

            if (not self.reset_apbootstrap_ini_sections()):
                print("apbootstrap: Could not reset apbootstrap sections (main.reset_apbootstrap_ini_sections)")
                sys.exit(1)

            if (not self.write_apbootstrap_ini()):
                print("apbootstrap: Could not write apbootstrap ini file (main.write_apbootstrap_ini)")
                sys.exit(1)

            sys.exit(0)

        # For normal operation use this path which runs through configuration if necessary
        else:
            if (not self.read_apbootstrap_ini()):
                print("apbootstrap: Could not read apbootstrap ini file (main.read_apbootstrap_ini)")
                sys.exit(1)

            if (not self.configure_apbootstrap_ini()):
                print("apbootstrap: Could not configure apbootstrap.ini file (main.configure_apbootstrap_ini)")
                sys.exit(1)

            print("apbootstrap: Should configure_initial_boot be executed?  Checking apbootstrap config:")
            if (self.check_if_section_should_be_executed("initial_boot")):
                if (not self.apbootstrap_initial_boot.configure_initial_boot()):
                    print("apbootstrap: Could not configure initial boot files (main.apbootstrap_initial_boot.configure_initial_boot)")
                    sys.exit(1)
                else:
                    if (not self.set_section_executed("initial_boot")):
                        print("apbootstrap: Could not set initial_boot executed in apbootstrap.ini (main.set_section_executed)")
                        sys.exit(1)

            print("apbootstrap: Should configure_autopilot_ini be executed?  Checking apbootstrap config:")
            if (self.check_if_section_should_be_executed("autopilot_ini")):
                if (not self.apbootstrap_autopilot_ini.configure_autopilot_ini()):
                    print("apbootstrap: Could not configure autopilot.ini file (main.apbootstrap_autopilot_ini.configure_autopilot_ini)")
                    sys.exit(1)
                else:
                    if (not self.set_section_executed("autopilot_ini")):
                        print("apbootstrap: Could not configure autopilot.ini file (main.set_section_executed)")
                        sys.exit(1)

            print("apbootstrap: Should configure_machine_env be executed?  Checking apbootstrap config:")
            if (self.check_if_section_should_be_executed("machine_env")):
                if (not self.apbootstrap_machine_env.configure_machine_env()):
                    print("apbootstrap: Could not configure machine ENV variables (main.apbootstrap_machine_env.configure_machine_env)")
                    sys.exit(1)
                else:
                    if (not self.set_section_executed("machine_env")):
                        print("apbootstrap: Could not configure machine ENV variables (main.set_section_executed)")
                        sys.exit(1)

            print("apbootstrap: Should configure_apt_proxy be executed?  Checking apbootstrap config:")
            if (self.check_if_section_should_be_executed("apt_proxy")):
                if (not self.apbootstrap_apt_proxy.configure_apt_proxy()):
                    print("apbootstrap: Could not configure apt proxy file (main.apbootstrap_apt_proxy.configure_apt_proxy)")
                    sys.exit(1)
                else:
                    if (not self.set_section_executed("apt_proxy")):
                        print("apbootstrap: Could not set apt_proxy executed in apbootstrap.ini (main.set_section_executed)")
                        sys.exit(1)

            ##########################
            # PARTITIONING GOES HERE #
            ##########################

            if (self.changed_apbootstrap_ini):
                self.write_apbootstrap_ini()

            # Exit in a happy :)  way!
            sys.exit(0)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Send a command to apbootstrap service.")
    # For resetting sections, run apbootstrap.py with -r "section1,section2,etc"
    # then re-run it again with no options to execute changes and set things back to Execute:False
    parser.add_argument("-r", "--reset", metavar="SECTION", help="section or sections to reset so they re-run next "
                                                                 "time apbootstrap runs; if supplying multiple "
                                                                 "sections please use commas to separate them and "
                                                                 "wrap them in double quotes")

    args = parser.parse_args()
    if (args.reset):
        sections_to_reset = args.reset.split(",")
    else:
        sections_to_reset = []

    apbootstrap_ini_path = apenv.get_data_root() + "/apbootstrap/apbootstrap.ini"
    apbootstrap_reboot_path = apenv.get_data_root() + "/apbootstrap/apbootstrap.reboot"
    linux_bootstrap_ini_path = apenv.get_linux_bootstrap_ini_path()
    autopilot_ini_path = apenv.get_autopilot_ini_path()
    global_environment_sh_path = "/etc/profile.d/ap.sh"
    apt_proxy_file = "/etc/apt/apt.conf.d/02proxy"
    apt_proxy_server = "http://aplinuxservices.trafficmanager.net:3142"
    initial_boot_csv_path = apenv.get_data_root() + "/apbootstrap/initial_boot.csv"
    initial_boot_csv_destination = apenv.get_data_root() + "/watchdogoutput/csv_sources/initial_boot.csv"
    watchdog_csv_dir = apenv.get_data_root() + "/watchdogoutput/csv_sources"

    Bootstrap(apbootstrap_ini_path,
              apbootstrap_reboot_path,
              linux_bootstrap_ini_path,
              autopilot_ini_path,
              global_environment_sh_path,
              apt_proxy_file,
              apt_proxy_server,
              sections_to_reset,
              initial_boot_csv_path,
              initial_boot_csv_destination,
              watchdog_csv_dir).main()
