#!/usr/bin/env python3

import configparser


# This class is used to remove spaces while writing an ini file to disk so that we can results of var=value instead
# of var = value
class INISpaceRemover:
    output_file = None

    def __init__(self, new_output_file):
        self.output_file = new_output_file

    def write(self, what):
        self.output_file.write(what.replace(" = ", "="))


class INIParser(object):
    # We specifically refuse to allow characters that cannot be parsed back in since the base extension doesn't
    # do a sanity check on set or write
    forbidden_characters = ["@", "%", "^", "`", ";", "#", "[", "]", "/", "\\", "//", "+", "="]

    def __init__(self, ini_path: str):
        self.ini_path = ini_path
        # print("\niniparser: Path passed to init: " + ini_path)

        # Initialize a new configparser object immediately
        self.config = configparser.ConfigParser()
        assert isinstance(self.config, object)
        # This option tells the parser to respect case of both var and value names
        # The normal behavior is to lowercase everything which is not wanted
        self.config.optionxform = str
        # print("iniparser: Initializing ConfigParser for: " + str(self.ini_path))

    # Create and return parser object
    def create(self) -> object:
        # print("iniparser: Creating new ini file in memory: " + self.ini_path)

        return self.config

    # Read existing file into parser object
    def read(self, ini_path: str="") -> bool:
        if(ini_path):
            self.ini_path = ini_path

        try:
            self.config.read(self.ini_path)

        except Exception as e:
            print("iniparser: Exception: " + str(e))
            return False

        return True

    # Return parser object
    def get_config_object(self) -> object:
        return self.config

    # Add a new section to the parser object
    def add_section(self, section: str) -> bool:
        # print("iniparser: Adding section: " + str(section))

        if ((type(section) != str)):
            print("iniparser: Cannot add section that is not a string!")
            return False

        if (set(list(section)).intersection(self.forbidden_characters)) or not section:
            print("iniparser: Section: " + section + " contains illegal character: " +
                  str(set(list(section)).intersection(self.forbidden_characters)))
            return False

        try:
            self.config.add_section(section)

        except Exception as e:
            print("iniparser: Exception: " + str(e))
            return False

        return True

    # Get a list of section names from the parser object
    def get_sections(self) -> list:
        return self.config.sections()

    # Return all sections and options as a dictionary of dictionaries
    def get_all_sections_and_options(self) -> dict:
        all_data = {}
        for each_section in self.config.sections():
            all_data[each_section] = {}
            for (each_key, each_val) in self.config.items(each_section):
                all_data[each_section][each_key] = each_val

        return all_data

    # Get an option from the parser object
    def get(self, section: str, option: str) -> str:
        result = self.config.get(section, option)
        # print("iniparser: Getting option: " + str(option) + " : " + str(result))

        return result

    # Get a boolean option from the parser object
    def getboolean(self, section: str, option: str) -> bool:
        result = self.config.getboolean(section, option)
        # print("iniparser: Getting boolean: " + str(option) + " : " + str(result))

        return result

    # Verify that a section exists in parser object
    def has_section(self, section: str) -> bool:
        if self.config.has_section(section):
            return True
        else:
            return False

    # Verify that an option exists in the parser object
    def has_option(self, section: str, option: str) -> bool:
        if self.config.has_option(section, option):
            return True
        else:
            return False

    # Set an option in the parser object
    def set(self, section: str, option: str, value: str) -> bool:
        # print("iniparser: Setting option: '" + str(option) + "' : '" + str(value) + "'")

        if ((type(option) != str) or (type(value) != str)):
            print("iniparser: Cannot add option or/with value that is not a string!")
            return False

        if not option or (set(list(option)).intersection(self.forbidden_characters)):
            print("iniparser: Variable: " + option + " contains illegal character: " +
                  str(set(list(option)).intersection(self.forbidden_characters)))
            return False

        if (set(list(value)).intersection(self.forbidden_characters)):
            print("iniparser: Variable: " + option + " Value: " + value + " contains illegal character: " +
                  str(set(list(value)).intersection(self.forbidden_characters)))
            return False

        try:
            self.config.set(section, option, value)

        except Exception as e:
            print("iniparser: Exception: " + str(e))
            return False

        return True

    # Remove an entire section in the parser object
    def remove_section(self, section: str) -> bool:
        # print("iniparser: Removing Section: " + str(section))
        try:
            self.config.remove_section(section)

        except Exception as e:
            print("iniparser: Exception: " + str(e))
            return False

        return True

    # Print the entire parser object contents in a nice structured view
    def print_all(self):
        print("")
        for each_section in self.config.sections():
            print(each_section)
            for (each_key, each_val) in self.config.items(each_section):
                print("\t" + str(each_key) + " : " + str(each_val))

    # Write the parser object to a file on disk
    def write(self) -> bool:
        # print("\niniparser: Writing ini file: " + self.ini_path)
        self.print_all()

        try:
            with open(self.ini_path, 'w', newline="\n") as inifile:
                self.config.write(INISpaceRemover(inifile))

        except Exception as e:
            print("iniparser: Exception: " + str(e))
            return False

        return True
