#!/usr/bin/env python3

# Class used to help send data to statsd
# Automatically includes things like dimensions to the output to make sending stats easier

import sys
import os
import socket
import json
import statsd

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv


class StatsdHelper(object):
    def __init__(self,
                 statsd_host: str="localhost",
                 statsd_port: int=9125,
                 test_data: dict={},
                 test_status: bool=False):

        self.test_status = test_status
        self.statsd_host = statsd_host
        self.statsd_port = statsd_port

        # print("host: " + str(self.statsd_host))
        # print("port: " + str(self.statsd_port))

        self.hostname = socket.gethostname()
        self.metric = statsd.StatsClient(statsd_host, statsd_port)

        # If this is a normal run we get our data from localwatchdogstatus.csv
        # and from apenv functions
        if not self.test_status:
            self.input_csv_file = apenv.get_data_root() + "/watchdogoutput/localwatchdogstatus.csv"

            # Tenant = AP Cluster
            # Role = AP Environment
            # RoleInstance = AP Machine Function
            self.tenant = apenv.get_cluster_name()
            self.role = apenv.get_environment_name()
            self.role_instance = apenv.get_machine_function()

        # If this is a unit test we get our data from test data passed in
        else:
            self.input_csv_file = test_data["input_csv_file"]
            self.tenant = test_data["Tenant"]
            self.role = test_data["Role"]
            self.role_instance = test_data["RoleInstance"]

        self.osversion = ""
        self.shared_services_version = ""

        # Call get_version_info method to get OS version and shared services version
        self.get_version_info()

    # Get version information for the OS and shared services
    def get_version_info(self) -> bool:
        # Verify that watchdog has already created this file
        if os.path.isfile(self.input_csv_file):
            try:
                with open(self.input_csv_file, 'r') as infile:
                    for line in infile:
                        if "OSImageVersion" in line:
                            self.osversion = line.split(",")[3].strip()
                        elif "LinuxSharedServicesVersion" in line:
                            self.shared_services_version = line.split(",")[3].strip()
            except Exception as e:
                print("statsdhelper.get_version_info: Exception: " + str(e))
                return False

            # If the values are not present yet in the file we can error out and try again later
            # once the watchapversion.py watchdog runs and gets this info
            if not self.osversion or not self.shared_services_version:
                return False
        else:
            return False

        return True

    # Send gauge statistic to statsd with automatically filled in Dimensions
    # Minimal input required to send a metric
    def send_gauge(self, namespace: str, metric_name: str, description: str, metric_value: int) -> bool:
        # print("\nSending statsd values:")
        # print("statsdhelper.send_gauge Namespace: " + namespace)
        # print("statsdhelper.send_gauge Metric: " + metric_name)
        # print("statsdhelper.send_gauge Description: " + description)
        # print("statsdhelper.send_gauge Tenant: " + self.tenant)
        # print("statsdhelper.send_gauge Role: " + self.role)
        # print("statsdhelper.send_gauge RoleInstance: " + self.role_instance)
        # print("statsdhelper.send_gauge MachineName: " + self.hostname)
        # print("statsdhelper.send_gauge OSVersion: " + self.osversion)
        # print("statsdhelper.send_gauge SharedServicesVersion: " + self.shared_services_version)
        # print("statsdhelper.send_gauge Value: " + str(metric_value))

        # Make sure all required dimensions are present and accounted for
        if not all((namespace, metric_name, metric_value, description, self.tenant, self.role, self.role_instance,
                    self.hostname, self.osversion, self.shared_services_version)):
            print("statsdhelper.send_gauge: Not all of the required dimensions were defined properly")

        metric_json = ""

        try:
            metric_dict = {"Namespace": namespace,
                           "Metric": metric_name,
                           "Dims": ""}

            dims_dict = {"Description": description,
                         "Tenant": self.tenant,
                         "Role": self.role,
                         "RoleInstance": self.role_instance,
                         "MachineName": self.hostname,
                         "OSVersion": self.osversion,
                         "SharedServicesVersion": self.shared_services_version}

            metric_dict["Dims"] = dims_dict

            metric_json = json.dumps(metric_dict)

            # Show JSON output sent to statsd
            # print("\nJSON Fromatted output for statsd:")
            # print(json.dumps(metric_dict, indent=4, sort_keys=True))

            self.metric.gauge(metric_json, metric_value)

        except Exception as e:
            print("statsdhelper.send_gauge: Exception: " + str(e))
            return False

        return True
