#!/usr/bin/env python3

import sys
import os
import stat
import argparse
import base64
from io import StringIO

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apcommon import apsecretclient


class GenevaSecretRefresher:
    SECRET_MDSD_APLINUXGENEVA_AUTOKEY = "Geneva-APLinuxGeneva-autokey"
    SECRET_MDSD_AZSECPACK_AUTOKEY = "Geneva-AzureSecurityPack-autokey"
    SECRET_MDSD_SYSLOG_AUTOKEY = "Geneva-Syslog-autokey"
    SECRET_MDSD_PFX = "Geneva-PrivateKeyPFX"
    SECRET_MDSD_PFX_PASS = "Geneva-PrivateKeyPFX-Password"

    SECRET_MDSDINT_APLINUXGENEVA_AUTOKEY = "GenevaINT-APLinuxGeneva-autokey"
    SECRET_MDSDINT_AZSECPACK_AUTOKEY = "GenevaINT-AzureSecurityPack-autokey"
    SECRET_MDSDINT_SYSLOG_AUTOKEY = "GenevaINT-Syslog-autokey"
    SECRET_MDSDINT_PFX = "GenevaINT-PrivateKeyPFX"
    SECRET_MDSDINT_PFX_PASS = "GenevaINT-PrivateKeyPFX-Password"

    def __init__(self, client_key_file, client_cert_file):
        self.key_file = client_key_file
        self.cert_file = client_cert_file
        self.kv_client = None

        self.mdsd_autokey = ""
        self.mdsd_pfx = ""
        self.mdsd_pfx_pass = ""

        # DC/Region Mappings for Geneva Warm Path Autokey Zones
        # See http://sharepoint/sites/autopilot/wiki/Virtual%20Machines.aspx
        self.autokey_regions = {}
        self.autokey_regions["CO4"] = "WUS2"  # uswest
        self.autokey_regions["BJ1"] = "WUS2"  # china
        self.autokey_regions["HK2"] = "WUS2"  # asia
        self.autokey_regions["BN1"] = "EUS"  # useast
        self.autokey_regions["BN2"] = "EUS"  # useast
        self.autokey_regions["BN2B"] = "EUS"  # useast
        self.autokey_regions["BN4C"] = "EUS"  # useast
        self.autokey_regions["CH1"] = "SCUS"  # usnorth
        self.autokey_regions["CH1B"] = "SCUS"  # usnorth
        self.autokey_regions["CH1D"] = "SCUS"  # usnorth
        self.autokey_regions["CO1C"] = "WUS2"  # uswest
        self.autokey_regions["CO3"] = "WUS2"  # uswest
        self.autokey_regions["CO3B"] = "WUS2"  # uswest
        self.autokey_regions["CO3C"] = "WUS2"  # uswest
        self.autokey_regions["CO4"] = "WUS2"  # uswest
        self.autokey_regions["CO4C"] = "WUS2"  # uswest
        self.autokey_regions["CY2"] = "SCUS"  # uscentral
        self.autokey_regions["CY2B"] = "SCUS"  # uscentral
        self.autokey_regions["CY2TEST01"] = "SCUS"  # uscentral
        self.autokey_regions["CY4"] = "SCUS"  # uscentral
        self.autokey_regions["DB3"] = "EUS"  # europe
        self.autokey_regions["DB4"] = "EUS"  # europe
        self.autokey_regions["DB5"] = "EUS"  # europe
        self.autokey_regions["DM2C"] = "SCUS"  # uscentral
        self.autokey_regions["DM3"] = "SCUS"  # uscentral
        self.autokey_regions["HK2"] = "WUS2"  # asia
        self.autokey_regions["MW1"] = "WUS2"  # uswest
        self.autokey_regions["SG1"] = "WUS2"  # asia
        self.autokey_regions["SN2"] = "SCUS"  # ussouth

        self.autokey_default_region = "WUS2"

    def __is_connected(self) -> bool:
        if self.kv_client is None:
            return False

        # TODO: Make this more resilient
        return True

    def __connect(self):
        if self.__is_connected():
            return
        self.kv_client = apsecretclient.APSecretClient(self.key_file, self.cert_file)

    # def __check_pfx_password(self):
    #     if self.mdsd_pfxpass == "":
    #         raise ValueError("PFX password empty or not set")
    #
    #     # TODO: Check to make sure password unlocks pfx

    def decode_base64_string(self, base64string: str) -> bytes:
        decoded_string = base64.b64decode(base64string, validate=True)

        return decoded_string

    def match_autokey_region(self, decoded_mdsd_autokey: str) -> str:
        # Get the current machine region from autopilot.ini
        machine_region = apenv.get_cluster_name().lower()

        # See whether the region of the machine exists in the default mappings at the top of this file
        # If not, assign to default region
        if machine_region.upper() in self.autokey_regions:
            current_autokey_region = self.autokey_regions[machine_region.upper()]
        else:
            current_autokey_region = self.autokey_default_region

        # Go through the autokey secrets file to see what the autokey is for the specified region
        string_object = StringIO(decoded_mdsd_autokey)
        current_autokey = ""

        for line in string_object:
            region = line.split(":")[0].lower()
            autokey = line.split(":")[1]

            if region == current_autokey_region.lower():
                current_autokey = autokey

        if not current_autokey:
            raise ValueError("Autokey for '{0}' Region does not exist".format(current_autokey_region))
        else:
            return current_autokey

    def build_autokey_string(self, current_autokey: str) -> str:
        autokey_text = "<config>\n  <bootstrapkeys>\n    <bootstrapkey keyPath=" + '"/etc/mdsd.d/privatekey.pfx" keyPass="' \
            + self.mdsd_pfx_pass + '">' + current_autokey.strip() + '</bootstrapkey>\n  </bootstrapkeys>\n</config>\n'

        return autokey_text

    def write_secret_string_file(self, directory: str, file_path: str, string_to_write: str):
        if not os.path.isdir(directory):
            raise ValueError("Download directory '{0}' does not exist".format(directory))
        else:
            try:
                    with open(file_path, "w", newline="\n") as text_file:
                        text_file.write(string_to_write)

            except Exception as e:
                print("apsecret_refresh_geneva_warm.write_secret_string_file: Exception: " + str(e))

    def write_secret_binary_file(self, directory: str, file_path: str, string_to_write: bytes):
        if not os.path.isdir(directory):
            raise ValueError("Download directory '{0}' does not exist".format(directory))
        else:
            try:
                    with open(file_path, "wb") as text_file:
                        text_file.write(string_to_write)

            except Exception as e:
                print("apsecret_refresh_geneva_warm.write_secret_string_file: Exception: " + str(e))

    def set_permissions_on_file(self, full_file_path: str):
        # Set secret file mode to 0600 so that no one can mess with this but root
        try:
            os.chown(full_file_path, 0, 0)
            os.chmod(full_file_path, stat.S_IRUSR | stat.S_IWUSR)

        except Exception as e:
            print("apsecret_refresh_geneva_warm.set_permissions_on_file: Exception: " + str(e))
            raise ValueError("Failed to set permissions on '{0}'".format(full_file_path))

    def download_geneva_certificates(self, directory: str, environment: str, namespace: str):
        # Determine the Environment and Namespace to use for certificates and folders based on which MDSD we are deploying
        self.__connect()
        if (environment == "prod"):
            if (namespace == "aplinuxgeneva"):
                self.mdsd_autokey = self.kv_client.read_cache_secret(self.SECRET_MDSD_APLINUXGENEVA_AUTOKEY)
            elif (namespace == "azsecpack"):
                self.mdsd_autokey = self.kv_client.read_cache_secret(self.SECRET_MDSD_AZSECPACK_AUTOKEY)
            elif (namespace == "syslog"):
                self.mdsd_autokey = self.kv_client.read_cache_secret(self.SECRET_MDSD_SYSLOG_AUTOKEY)
            else:
                raise ValueError("Namespace not specified and it is required")

            self.mdsd_pfx = self.kv_client.read_cache_secret(self.SECRET_MDSD_PFX)
            self.mdsd_pfx_pass = self.kv_client.read_cache_secret(self.SECRET_MDSD_PFX_PASS)

        elif (environment == "int"):
            if (namespace == "aplinuxgeneva"):
                self.mdsd_autokey = self.kv_client.read_cache_secret(self.SECRET_MDSDINT_APLINUXGENEVA_AUTOKEY)
            elif (namespace == "azsecpack"):
                self.mdsd_autokey = self.kv_client.read_cache_secret(self.SECRET_MDSDINT_AZSECPACK_AUTOKEY)
            elif (namespace == "syslog"):
                self.mdsd_autokey = self.kv_client.read_cache_secret(self.SECRET_MDSDINT_SYSLOG_AUTOKEY)
            else:
                raise ValueError("Namespace not specified and it is required")

            self.mdsd_pfx = self.kv_client.read_cache_secret(self.SECRET_MDSDINT_PFX)
            self.mdsd_pfx_pass = self.kv_client.read_cache_secret(self.SECRET_MDSDINT_PFX_PASS)
        else:
            raise ValueError("Environment not specified and it is required")

        # Decode Secrets from base64 to normal
        decoded_mdsd_autokey = self.decode_base64_string(self.mdsd_autokey).decode('utf-8')
        # PFX needs to be left in binary mode because it's a binary file even after decoding
        decoded_mdsd_pfx = self.decode_base64_string(self.mdsd_pfx)

        # Match Autokey Region
        current_autokey = self.match_autokey_region(decoded_mdsd_autokey)

        # Build Autokey File
        formatted_autokey_string = self.build_autokey_string(current_autokey)

        # Make sure the download directory exists. If it does not, then we were
        # just refreshing the secret cache
        if not os.path.isdir(directory):
            return

        # Write Autokey File
        autokey_file_path = os.path.join(directory, "mdsautokey.cfg")
        self.write_secret_string_file(directory, autokey_file_path, formatted_autokey_string)
        self.set_permissions_on_file(autokey_file_path)

        # Write PFX File in binary mode
        pfx_file_path = os.path.join(directory, "privatekey.pfx")
        self.write_secret_binary_file(directory, pfx_file_path, decoded_mdsd_pfx)
        self.set_permissions_on_file(pfx_file_path)


def main(geneva_data_directory: str, environment: str, namespace: str):
    client_key, client_cert = apenv.get_machine_function_cert()
    refresher = GenevaSecretRefresher(client_key, client_cert)

    try:
        refresher.download_geneva_certificates(geneva_data_directory, environment, namespace)

    except Exception as e:
        print("ERROR: {0}".format(e), file=sys.stderr)
        sys.exit(1)

    sys.exit(0)

# Example run command:
# /ap/app/aptools/apsecret_refresh_geneva_warm.py -d /ap/data/geneva/mdsd_azuresecuritypack -e prod -n azsecpack

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Refreshes secrets for accessing Geneva Warm Path and copies "
                                     "certificates to the specified directory.")
    parser.add_argument("-d", "--download", metavar="DIRECTORY", help="Directory to copy certificates into")
    parser.add_argument("-e", "--environment", metavar="ENVIRONMENT", help="Environment to use - prod/int")
    parser.add_argument("-n", "--namespace", metavar="NAMESPACE", help="Geneva Namespace to use aplinuxgeneva/azsecpack/syslog")

    args = parser.parse_args()

    if args.environment is None or args.environment.strip() not in ["prod", "int"]:
        print("ERROR: Environment must be specified. Valid options are: prod int", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    if args.namespace is None or args.namespace.strip() not in ["aplinuxgeneva", "azsecpack", "syslog"]:
        print("ERROR: Namespace must be specified. Valid options are: aplinuxgeneva azsecpack syslog", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    main(args.download, args.environment, args.namespace)
