#!/usr/bin/env python3
# GetAndRunServices equivalent in Linux on AP.
# Scheduled once every 5 mins
# 1. Gets the Service List from DM
# 2. Gets ServiceList ini from manifest server if there is a new version for
#    deployment
# 3. If PrepareList field has changed from last time, download the package from DS Server.
#     - Using HTTP
# 4. If 'List' changed, then inform ServiceManager to start the new versions of services.
# 5. Drop the new servicelist.json file for service manager
# 6. Update DM with the new service list version, even if no changes happened, but more the
#    6hrs passed since last update

import argparse
from configparser import ConfigParser, Error
from json import load, dump
from sys import path as syspath
from os import path, rename, remove
from shutil import os
from tempfile import NamedTemporaryFile
from requests import codes, get, exceptions
from statsd import StatsClient
from socket import gethostname
from time import sleep
from typing import DefaultDict, List, Union
from logging import getLogger, config

# if script's parent directory isn't in sys.path, then add it
if '/'.join(syspath[0].split("/")[:-1]) not in syspath:
    syspath.append('/'.join(syspath[0].split("/")[:-1]))

from apsdk import apenv
from apsdk import dmclient
from apcommon import apck

# globals
# Create a metric instance to upload metrics
# Follow convention when metrics:
#  error = error.<dot-delimited-path>. e.g. error.dm.GetServiceList.timeout
#  performance = perf.<dot-delimited-path>
metric = StatsClient(host='localhost', port=9125, prefix='gars')

# New Service list JSON file that GARS writes for informing service manager of List it
# should be running
new_json_file = path.join(apenv.get_data_root(), 'servicemanager/new_servicelist.json')

# Active Service list JSON file that Service Manager writes for informing GARS of current List
# it was successfully able to activate
active_json_file = path.join(apenv.get_data_root(), 'servicemanager/active_servicelist.json')

# All the service lists are archived in a servicelists folder
service_list_dir = path.join(apenv.get_app_root(), 'servicelists')

# Packed images directory contains all the images unpacked by GARS
packed_images_dir = path.join(apenv.get_app_root(), 'packedimages')

copy_only_header = '#CopyOnly: '
data_copy_header = '#Data: '

# Csv to ignore when we process the service list
ignore_csv = '.Autopilot-AutopilotClient-VE.csv'

# Setup logger
config.fileConfig(path.join(syspath[0], 'logging.ini'))
log = getLogger(__name__)


def get_current_list(ini_config: ConfigParser) -> str:
    return ini_config['ServiceManager']['List']


def get_prepare_list(ini_config: ConfigParser) -> str:
    return ini_config['ServiceManager']['PrepareList']


def send_dm_command(dm_client: dmclient.DMClient,
                    cmd: str, retries=4) -> str:
    """
    Send a command to DM. Handles retries as sometimes DM doesn't respond with
    proper response.
    """
    for i in range(0, retries):
        resp = dm_client.send_request(cmd, '')
        if resp is not None:
            reply = resp.text
            resp.connection.close()
            return reply

        sleep(2)

    log.error("Sending command to DM failed")
    return None


def get_service_list_from_dm(dm_client: dmclient.DMClient) -> str:
    """
    Gets the service list from Device Manager
    """
    machine_name = gethostname()
    env_name = apenv.get_environment_name()

    cmd = 'GetServiceList -s ' + machine_name + ' -w ' + env_name

    log.info("Getting service list from DM: %s", cmd)
    with metric.timer('perf.dm.cmd.GetServiceList.timing'):
        resp = send_dm_command(dm_client, cmd)

    return resp


def update_dm_service_status(dm_client: dmclient.DMClient,
                             current_list: str,
                             prepare_list: str) -> str:
    """
    Update DM with current service status
    """
    # Command to update service list to DM
    cmd = 'ServiceStatus -serviceList ' + current_list + ' -prepareServiceList ' + prepare_list

    log.info("Updating service status to DM: %s", cmd)
    with metric.timer('perf.dm.cmd.ServiceStatus.timing'):
        resp = send_dm_command(dm_client, cmd)

    return resp


def get_manifest_server_list(ini_config: ConfigParser) -> list:
    """
    Get the manifest server list from an ini config returned
    """
    if not ini_config.has_section('Manifest'):
        log.error('DM Reply does not have [Manifest] section : %s', ini_config)
        return None
    else:
        server_list = ini_config['Manifest']['ServerList']
        return server_list.split(':')


def get_env_name_from_svc_list_name(service_list_name: str,
                                    default_env='') -> str:
    """
    Get environment name from a service list name
    """
    service_list_suffix = extract_service_list_suffix(service_list_name, False)

    # Dynamic service lists look like virtual ones, but they belong to
    # the current environment, not VE
    if service_list_suffix is not None and service_list_suffix.lower() == 'dynamic':
        return default_env

    return service_list_suffix


def get_pkgs_from_service_list(service_list_csv_path: str) -> List[str]:
    """
    Get package list from the service list specified
    """
    pkg_list = list()
    try:
        with open(service_list_csv_path) as f:
            for line in f:
                if line.startswith('#'):
                    # if line starts with #, we don't care about it yet
                    continue
                else:
                    # Line is of format:
                    #  APAMLocal.az_compute_2065832_AutoDeployment,0,0,,
                    # Get just the package name out of it
                    items = line.split(sep=',')[0]
                    if len(items) > 4:
                        pkg_name = line.split(sep=',')[0]
                        pkg_list.append(pkg_name)
                    else:
                        # We ignore this error since some lines are not parsable
                        log.debug("Package name is not of expected format: %s, in service list: %s",
                                  line,
                                  service_list_csv_path)

        return pkg_list

    except IOError as e:
        log.error("Service List csv could not be read : %s, error %s", service_list_csv_path, e)
        return None


def extract_service_list_suffix(service_list_name: str,
                                bootstrap=False) -> str:
    """
    Extract the service list suffix
    """
    if bootstrap:
        return 'bootstrap'

    items = service_list_name.split('.')
    count = len(items)
    if (count > 3) and items[count - 1] == 'csv':
        suffix = items[count - 2]
    else:
        suffix = None

    return suffix


def get_apck_url(ip: str,
                 listen_port: str,
                 env_name: str,
                 service_name: str) -> str:
    url = str.join('', ('http://', ip, ':', str(listen_port),
                        '//DynamicStorage/?action=DownloadStream&StreamName=stream://',
                        env_name, '/app/app.retail.amd64/',
                        service_name, '&version=apck'))
    return url


def get_manifest_server_url(server: str,
                            port: str,
                            environment: str,
                            csv_name: str) -> str:
    url = str.join('', ('http://', server, ':', port,
                        '/ManifestServer?action=GetServiceList&environment=',
                        environment, '&name=', csv_name))
    return url


def download_service_list_csv(manifest_servers: List[str],
                              listen_port: str,
                              environment: str,
                              csv_name: str,
                              timeout: int) -> str:
    """
    Download the service list csv file from manifest server
    """
    # Try from each server in the manifest server list
    for server in manifest_servers:
        url = get_manifest_server_url(server, str(listen_port), environment, csv_name)

        # We will get a "301 - Moved Permanently" from Manifest server that
        # redirects us to the actual DynamicStorage url
        try:
            log.info("Download Service List : %s", url)
            resp = get(url, allow_redirects=True, timeout=timeout)
            if resp.status_code == codes.ok:
                content = resp.text
                resp.close()
                log.debug("Response : %s", content)
                return content
            else:
                log.error("Error downloading service list: resp.Stats=%s, resp=%s",
                          resp.status_code, resp.text)

        except exceptions.ConnectTimeout:
            log.error("ConnectTimeout: %s", url)
        except exceptions.ConnectionError:
            log.error("ConnectionError: %s", url)
        except exceptions.Timeout:
            log.error("Timeout: %s", url)

    return None


@metric.timer('perf.dm.download_package')
def download_package(service_name: str,
                     service_list_name: str,
                     image_root: str,
                     manifest_servers: List[str],
                     listen_port: str,
                     timeout=20) -> str:
    """
    Download packages in a service list.
    """
    env_name = get_env_name_from_svc_list_name(service_list_name)

    # If this is a physical environment
    if env_name is None:
        env_name = apenv.get_environment_name()

    # Create a directory with service_list_name to put the apck files
    service_file_name = path.basename(service_list_name)

    # Filename may start with a '.'
    target_dir = path.join(image_root, service_file_name)
    if not path.exists(target_dir):
        try:
            os.mkdir(target_dir)
        except OSError as e:
            log.error("Creating directory: %s. error=%s", target_dir, str(e))
            return None

    # Try downloading from each manifest server in the list
    for ip in manifest_servers:
        apck_url = get_apck_url(ip, listen_port, env_name, service_name)
        log.debug("Downloading Service Pkg from : %s", apck_url)

        # We will get a file to download from DynamicStorage server
        try:
            # Download as a stream as the size might become very large.
            # Stream download ensures memory doesn't blow out, also lets us
            # separate handle header download and content download separately
            with metric.timer('perf.pkg.download'):
                resp = get(apck_url,
                           stream=True,
                           allow_redirects=True,
                           timeout=timeout)

            if resp.status_code != codes.ok:
                log.debug("Error reading pkg file. resp.StatusCode=%s, resp=%s. Trying next server.",
                          resp.status_code, resp.text)
                resp.close()
                continue

            tmp_file_name = write_tmp_pkg(resp, service_name, target_dir)
            if tmp_file_name is None:
                log.error("Couldn't download file from url: %s", apck_url)
                return

            # Everything is fine, rename to original apck file name
            target_file_name = str.join('', (target_dir, '/', service_name, '.apck'))
            try:
                rename(tmp_file_name, target_file_name)
            except IOError as e:
                log.error("Renaming package from %s to %s. error=%s",
                          tmp_file_name, target_file_name, str(e))
                return None

            return target_file_name

        except exceptions.ConnectTimeout:
            log.error("ConnectTimeout: %s", apck_url)
            continue
        except exceptions.ConnectionError:
            log.error("ConnectionError: %s", apck_url)
            continue
        except exceptions.Timeout:
            log.error("Timeout: %s", apck_url)
            continue

    return None


def write_tmp_pkg(resp, service_name: str,
                  target_dir: str) -> str:
    # Download the file to a tmp file and rename if all looks well
    file_tmp = NamedTemporaryFile(mode='wb',
                                  dir=target_dir,
                                  prefix=service_name,
                                  suffix='.downloading',
                                  delete=False)
    # Handle file download now.
    chunk_size = 1024 * 5
    for chunk in resp.iter_content(chunk_size):
        if chunk:
            file_tmp.write(chunk)
            file_tmp.flush()

    # release the connection back to the pool
    resp.close()

    download_size = file_tmp.tell()
    file_tmp.close()

    # If we didn't receive the full file
    # TODO: check CRC
    # if download_size == 0 or download_size != resp.headers['X-FileSize']:
    if download_size == 0:
        log.error("Error: Didn't receive full file. expected:%d, Downloaded:%d",
                  download_size)
        remove(file_tmp.name)
        return None

    return file_tmp.name


def parse_dm_reply(dm_reply: str) -> Union[DefaultDict, List]:
    """
    Parse the DM reply. Will be in Ini format.
    """
    try:
        reply = ConfigParser()
        reply.read_string(dm_reply)
    except Error as e:
        log.error("Reading DM reply: %s. Error=%s", dm_reply, str(e))
        return None

    # Verify we got the mandatory fields needed
    sections = ['ServiceManager', 'Manifest']
    for section in sections:
        if not reply.has_section(section):
            log.error('DM Reply does not have [%s] section:%s',
                      section,
                      dm_reply)
            return None

    # Get the ServiceManager section
    service_manager_dict = reply['ServiceManager']

    # Verify we got the mandatory fields
    fields = ['List', 'PrepareList', 'LastList', 'LastPrepareList']
    for field in fields:
        if field not in service_manager_dict:
            log.error(
                'DM Reply does not have field [ServiceManager] %s in reply: ',
                dm_reply)
            return None

    if 'ServerList' not in reply['Manifest']:
        log.error('DM Reply does not have [Manifest].ServerList %s:',
                  dm_reply)
        return None

    # Get the IPs of manifest servers
    manifest_servers = reply['Manifest']['ServerList'].split(':')

    return service_manager_dict, manifest_servers


def download_pkgs_in_csv(csv_file: str,
                         manifest_servers: List[str]) -> List:
    """
    Download the packages listed in csv
    """
    log.debug("Downloading packages from : %s", csv_file)
    if csv_file.endswith(ignore_csv):
        log.info("Ignoring csv file : %s. Matches pattern: %s", csv_file, ignore_csv)
        return None

    pkg_list = get_pkgs_from_service_list(csv_file)
    if pkg_list is None:
        log.error("Package list couldn't be obtained.")

    if not path.isdir(packed_images_dir):
        try:
            os.mkdir(packed_images_dir)
        except OSError as e:
            log.error("Creating directory: %s. error=%s", packed_images_dir, str(e))
            return None

    downloaded_packages = list()
    for pkg_name in pkg_list:
        log.debug("Downloading Package: %s", pkg_name)
        apck_file_name = download_package(pkg_name,
                                          csv_file,
                                          packed_images_dir,
                                          manifest_servers,
                                          apenv.get_fsserver_http_port(),
                                          timeout=20)
        if apck_file_name is None:
            metric.incr('error.pkgs.downloaded')
            log.error("Error Could not download : %s", pkg_name)
        else:
            downloaded_packages.append(apck_file_name)
            metric.incr('pkgs.downloaded')
            log.debug("Downloaded apck : %s", apck_file_name)

    return downloaded_packages


def download_csv_from_list(csv_list: str,
                           manifest_servers: List[str]) -> List[str]:
    """
    Download the CSV files containing list of packages from a list sent
    """
    log.debug("Downloading CSVs from: %s", csv_list)

    if not path.isdir(service_list_dir):
        try:
            os.mkdir(service_list_dir)
        except OSError as e:
            log.error("Creating directory: %s, error=%s", service_list_dir, str(e))
            return None

    # Each list contains a set of csv file names delimited by a :
    csv_list = csv_list.split(':')
    csv_file_list = list()

    for csv_name in csv_list:
        # Special case fo OM - Orchestration manager. We are going to ignore
        # it for now. We will tango with OM later.
        if csv_name.startswith('.OM.'):
            continue

        csv_content = download_service_list_csv(manifest_servers,
                                                apenv.get_fsserver_http_port(),
                                                apenv.get_environment_name(),
                                                csv_name,
                                                timeout=15)
        if csv_content is None:
            log.error("Error reading csv from manifest server : %s", csv_name)
            return None

        # Write csv contents to a file
        csv_full_path = path.join(service_list_dir, csv_name)
        try:
            with open(csv_full_path, 'w') as f:
                csv_file_list.append(f.name)
                f.write(csv_content)
        except IOError as e:
            log.error("Error writing csv content to file: %s. error=%s",
                      csv_full_path, str(e))
            return None

    return csv_file_list


@metric.timer('perf.pkg.extract')
def extract_service_package(apck_file_name: str) -> bool:
    """
    Extract the apck package sent into the same directory
    """
    if not path.exists(apck_file_name):
        log.error("APCK file not found: %s", apck_file_name)
        return False

    try:
        apck_reader = apck.ApPackReader(apck_file_name)
    except ValueError as e:
        log.error("APCK file parsing failed: %s", str(e))
        return False
    except OSError as e:
        log.error("Creating an ApPackReader failed with exception: %s", str(e))
        return False

    # Create the package directory. Package directory is
    # /ap/app/packedimages/<pkg-name>/
    name_without_ext = path.splitext(path.basename(apck_file_name))[0]
    target_dir = path.join(path.dirname(apck_file_name), name_without_ext)
    if not path.isdir(target_dir):
        try:
            os.mkdir(target_dir)
        except OSError as e:
            log.error("Creating directory: %s. error=%s", target_dir, str(e))
            return False

    # Extract all files in the apck file in the target directory
    try:
        log.debug("Extracting package : %s", apck_file_name)
        apck_reader.extract_all(target_dir)
        log.debug("Extracting package Complete : %s", apck_file_name)
    except (OSError, IOError) as e:
        log.error("Extracting Package %s in directory: %s. error=%s",
                  apck_file_name, target_dir, str(e))
        return False

    return True


def process_prepare_list(service_manager_dict,
                         manifest_servers: List[str],
                         force_install: bool) -> str:
    """
    If PrepareList changes, proceed downloading the packages. This is the
    Pre-Fetching mechanism used by DM/GARS. We will later get a List entry
    change that will show which service list should service manager be running.
    """
    last_prepare_list = service_manager_dict['LastPrepareList']
    prepare_list = service_manager_dict['PrepareList']

    if prepare_list == last_prepare_list:
        if not force_install:
            log.info('PrepareList version did not change: %s', prepare_list)
            return prepare_list

        log.info('PrepareList version did not change but forcing install anyways: %s', prepare_list)

    # Download the CSV files
    downloaded_csv_files = download_csv_from_list(prepare_list, manifest_servers)
    if downloaded_csv_files is None or len(downloaded_csv_files) is 0:
        log.debug('No CSVs were downloaded. No valid CSVs for Linux.')
        return prepare_list

    # Download the packages listed in the csv file
    pkg_lists_downloaded = []
    for csv_file in downloaded_csv_files:
        if csv_file.endswith(ignore_csv):
            log.info("Ignoring csv file : %s. Matches pattern: %s", csv_file, ignore_csv)
            continue

        downloaded_pkg_list = download_pkgs_in_csv(csv_file, manifest_servers)
        if downloaded_pkg_list is None:
            log.error('APCK Packages could not be downloaded for : %s', csv_file)
            return None

        pkg_lists_downloaded.append(downloaded_pkg_list)
        log.debug("%d packages downloaded. List: %s", len(downloaded_pkg_list), downloaded_pkg_list)

    # Extract all the downloaded APCK packages.
    for package_list in pkg_lists_downloaded:
        for package_path in package_list:
            if extract_service_package(package_path) is False:
                log.error("Extracting package %s", package_path)
                return None
            # Remove apck file once extracted
            log.debug("Removing apck file after extraction: %s", package_path)
            os.remove(package_path)

    return prepare_list


def inform_service_manager(service_manager_config: DefaultDict) -> None:
    """
    Drops a file for service manager to pick up and switch service list
    """
    # Object returned by the config parser is not a dictionary, though
    # it looks like one, hence isn't JSON serializable. Put it in an object
    # so that we can serialize
    json_serializable = {
        'List': service_manager_config['List'],
        'LastList': service_manager_config['LastList'],
        'PrepareList': service_manager_config['PrepareList'],
        'LastPrepareList': service_manager_config['LastPrepareList']
    }

    try:
        fp = open(new_json_file, 'w')
    except IOError as e:
        log.error("Could not open file for writing: %s. error=%s", new_json_file, str(e))
        return

    with fp:
        try:
            dump(json_serializable, fp, indent=2)
        except TypeError as e:
            log.error("JSON dump error: %s", str(e))
            return


def read_active_list(file_name: str) -> DefaultDict:
    """
    Read the active_list json and return the dictionary
    """
    if not path.exists(file_name):
        log.info('active_servicelist.json not found:%s', active_json_file)
        return None

    try:
        fp = open(file_name, 'r')
    except IOError as e:
        log.error("Could not read file %s. error=%s", file_name, str(e))
        return None

    with fp:
        try:
            return load(fp)
        except ValueError as e:
            log.error("JSON syntax error: %s", str(e))

    return None


def report_to_dm(dm_client: dmclient.DMClient,
                 prepare_list: str) -> bool:
    """
    Report current service status to Device Manager
    """

    # Read the current list json file
    active_dict = read_active_list(active_json_file)

    if active_dict is None:
        log.info('active_servicelist.json not found or could not be read:%s', active_json_file)
        return False

    # List should come out of active_servicelist.json since this is written by
    # service manager and it does so only when it can process the packages
    # specified in the list.
    log.debug("Updating service status to DM")
    resp = update_dm_service_status(dm_client, active_dict['List'], prepare_list)

    if resp is None:
        log.error('Updating DM Service Status')
        return False

    if not resp.startswith('OK'):
        log.error('DM did not respond with OK to Service Status update : %s', resp)

    return True


def main(force_install: bool):
    # Get a new DM Client to talk to the lord(DM)
    dm_client = dmclient.DMClient('')

    # Get service list from Device Manager
    service_list_reply = get_service_list_from_dm(dm_client)
    if service_list_reply is None:
        log.error("Getting service list from DM failed.")
        exit(0)
    log.debug("DM Reply: %s", service_list_reply)

    if service_list_reply.startswith('No Response'):
        log.error("No Reponse received from DM. This is a valid response.")
        exit(0)

    # Parse DM service list reply and obtain values to work with.
    service_manager_dict, manifest_servers = parse_dm_reply(service_list_reply)
    if service_manager_dict is None or manifest_servers is None:
        log.error("Could not parse ServiceList ini returned by DM.")
        exit(0)

    # Process the PrepareList entry in the Service Ini. PrepareList indicates
    # that GARS should pre-fetch the packages.
    processed_prepare_list = process_prepare_list(service_manager_dict, manifest_servers, force_install)

    if processed_prepare_list is None:
        log.error("Processing prepare list failed. Cannot proceed.")
        exit(1)

    # If we successfully downloaded the packages, we should inform
    # service manager to process. Should do this regardless of PrepareList being
    # new or old, since the 'List' entry is what matters to service manager.
    inform_service_manager(service_manager_dict)

    # Send List and PrepareList back to DM informing the versions this machine is at.
    if report_to_dm(dm_client, processed_prepare_list) is False:
        log.error("Reporting to DM failed. Exitting GARS")
        exit(0)

    log.info("GARS processing completed.")

    exit(0)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Send a command to getandrunservices service.")
    parser.add_argument("-f", "--force", action='store_true', help="force download and install of current servicelist")

    args = parser.parse_args()

    main(args.force)
