#!/usr/bin/env python
# Copyright (c) Microsoft Corporation.  All rights reserved.

"""
    APCAClient
        Parameters:
            -n  -- Name of the device requesting the certificate
            -i  -- Local interface IP address to bound to (optional, default=any)
            -s  -- APCA service address (optional, default=APCA-backend.Autopilot.Co3c.ap.gbl:443)
            -d  -- If specified will run as a daemon (optional, default=false)
            -r  -- If specified will try to renew only (optional, default=false)
            -t  -- Specifies the seconds to sleep between cycles when running as deamon (optional, default=600)
            -w  -- Specifies the number of days of early expiration (optional, default=12)
            -e  -- Perform AP PKI CA certificates retrieval. This is normally used for bootstrapping.
            -m  -- Request a MachineFunction Cert from APCA. Should supply dir containing machine certs.
            -f  -- Specifies the directory where certificates and keys will be read or written (optional, default=.)

    Examples:
        -   Retrieve, for machine CH1SWANBORDER01, the AP PKI CA certificates and place them in the subfolder ./cas
            Use interface 10.11.151.32 and APCA service at 65.55.106.202
                > python apca_client.py -n CH1SWANBORDER01 -e -i 10.11.151.32 -s 65.55.106.202 -f cas

        -   Retrieve Machine Function Cert. Need to pass dir containing machine certs obtained.
                > python apca_client.py -n CH1SWANBORDER01 -e -i 10.11.151.32 -s 65.55.106.202 -m mf_certs -f machine_certs

        -   Run as a deamon for machine CH1SWANBORDER01 and checks every 30 seconds. If a certificate exists
            for the local machine the just check for renewals. If certificate does not exist then provision a new one.
            Stores all the certificates and keys in the ./localCerts directory
            Use interface 10.11.151.32 and APCA service at 65.55.106.202
                > python apca_client.py -n CH1SWANBORDER01 -i 10.11.151.32 -s 65.55.106.202 -d -t 30 -f localCerts
"""
import urllib
import sys
import getopt
import os
import hashlib
import httplib
import re
import socket
import ssl
import time
import subprocess
from datetime import datetime, timedelta
from M2Crypto import RSA, X509, EVP, m2

# Do not allow to import all symbols when using "from apca_client import *"
__all__ = []

# Globals
local_ip_address = None
rsa_key_bits = 2048
pem_req_header = "-----BEGIN CERTIFICATE REQUEST-----"
pem_req_trailer = "-----END CERTIFICATE REQUEST-----"
pem_cert_header = "-----BEGIN CERTIFICATE-----"
pem_cert_trailer = "-----END CERTIFICATE-----"
cert_dir = os.path.dirname(os.path.realpath(__file__))

apca_url = "APCA-backend.Autopilot.Co3c.ap.gbl:443"
apca_new_cert_ep = "/APCAOneWay/SignMachineCertInitial"
apca_renew_cert_ep = "/APCAMutual/SignMachineCertRenew"
apca_export_pki_ep = "/APCAOneWay/ExportApPkiStore"
apca_new_mf_cert_ep = "/APCAMutual/SignMachineFunction"


class HTTPSClient(httplib.HTTPSConnection):
    """ Class to make a HTTPS connection, with support for binding to iface """

    def __init__(self,
                 host,
                 timeout=None,
                 local_ip=None,
                 keyfile=None,
                 certfile=None):
        httplib.HTTPSConnection.__init__(self, host, timeout=timeout)
        self.local_ip = local_ip
        self.key_file = keyfile
        self.cert_file = certfile

    def connect(self):
        """ Connect to a host on a given (SSL) port. """
        sock = None
        if (self.local_ip is None):
            print "Connecting to %s on port %s" % (self.host, self.port)
            sock = socket.create_connection((self.host, self.port),
                                            self.timeout)
        else:
            sock = socket.create_connection((self.host, self.port),
                                            self.timeout,
                                            (self.local_ip, 0))

        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        self.sock = ssl.wrap_socket(sock,
                                    self.key_file,
                                    self.cert_file,
                                    cert_reqs=ssl.CERT_NONE)


def cdir_path(path):
    """ Returns the current output path """
    return os.path.join(cert_dir, path)


def generate_rsa_key():
    """ Creates an asymmetric key """
    return RSA.gen_key(rsa_key_bits, m2.RSA_F4)


def make_private_key(key):
    """ Creates a private key """
    pkey = EVP.PKey()
    pkey.assign_rsa(key)
    return pkey


def get_public_key_fingerprint(keypair):
    digest = hashlib.sha1(keypair.as_der()).hexdigest().upper()
    return ":".join(digest[pos: pos + 2] for pos in range(0, 40, 2))


def make_cert_request(pkey, server_dns):
    """ Creates a X509 certificate sign request """
    req = X509.Request()
    req.set_version(req.get_version())

    # set public key
    req.set_pubkey(pkey)

    # set CN field
    name = X509.X509_Name()
    name.CN = server_dns
    req.set_subject(name)

    # Extensions array of tuples (name, value, if_critical)
    extensions = [("extendedKeyUsage", "serverAuth,clientAuth", False),
                  ("keyUsage", "dataEncipherment, keyEncipherment", True),
                  ("subjectKeyIdentifier", get_public_key_fingerprint(pkey), False)]

    # Set extensions
    extstack = X509.X509_Extension_Stack()

    for extension in extensions:
        ext = X509.new_extension(extension[0], extension[1])
        if (extension[2]):
            ext.set_critical()
        extstack.push(ext)

    req.add_extensions(extstack)

    # add SHA1 signature
    req.sign(pkey, "sha1")

    return req


def apca_http_request(conn, url, method="POST", postData=None):
    """ Perform HTTP request """
    headers = {"content-type": "text/plain"}

    print "\t> %s HTTP request to APCA %s" % (method, url)

    conn.request(method, url=url, body=postData, headers=headers)
    response = conn.getresponse()
    respbody = response.read()

    print "\t> ", response.status, response.reason
    return response.status, response.reason, respbody


def apca_http_post(conn, url, postData):
    """ Perform HTTP post request """
    return apca_http_request(conn, url, "POST", postData)


def apca_http_get(conn, url):
    """ Perform HTTP get request """
    return apca_http_request(conn, url, "GET")


def apca_get_querystring(machine, caThumbprint=""):
    """ Returns the query string of parameters as requested by
        APCA REST api"""
    params = {"useAppkiV2": "true",
              "useShortLifetime": "false",
              "requestedMachine": machine,
              # "requestedMachineFunctionFqdns": "APTest04.Autopilot.AzureCloudvSwitch",
              "requestedMachineFunctionFqdns": "",
              "requestedCAThumbprint": caThumbprint}
    return "?" + urllib.urlencode(params)


def remove_pem_req_header_trailer(cert):
    """ Helper that removes header and trailer from PEM certs """
    return cert.replace(pem_req_header, "").replace(pem_req_trailer, "")


def format_pem_cert_header_trailer(cert):
    """ Helper that appends header and trailer to base64 PEM certs
        with appropriate formatting """
    cert = re.sub("(.{64})", "\\1\n", cert, 0, re.DOTALL)
    if not cert.startswith(pem_cert_header):
        return pem_cert_header + "\n" + cert.strip() + "\n" + pem_cert_trailer
    return cert


def apca_new_cert(crtreq, uri, machine, local_ip=None):
    """ Request APCA to sign a new certificate by a CSR """
    print "Povisioning a new certificate"
    # send to a web service to sign the certificate
    conn = HTTPSClient(uri, timeout=5, local_ip=local_ip)

    url = apca_new_cert_ep + apca_get_querystring(machine)

    # removing pre and post strings from certificate request
    crtreq = remove_pem_req_header_trailer(crtreq)
    postData = unicode(crtreq, "utf-8")

    # perform request
    status, reason, body = apca_http_post(conn, url, postData)

    # close connection
    conn.close()

    if (status == 200):
        return format_pem_cert_header_trailer(body)

    raise Exception(str(status), body)


def apca_new_mf_cert(crtreq, uri, machine, local_ip=None, machine_dir=""):
    """ Request APCA to renew an existing certificate by a CSR """
    print "Obtaining a MACHINE FUNCTION Cert using Machine Certs in " + machine_dir
    # send to a web service to sign the certificate
    conn = HTTPSClient(uri,
                       timeout=5,
                       local_ip=local_ip,
                       # keyfile="./cert/"+machine+".key",
                       # certfile="./cert/"+machine+".crt")
                       keyfile=machine_dir + "/" + machine + ".key",
                       certfile=machine_dir + "/" + machine + ".crt")

    url = apca_new_mf_cert_ep + apca_get_querystring(machine)

    # removing pre and post strings from certificate request
    crtreq = remove_pem_req_header_trailer(crtreq)
    postData = unicode(crtreq, "utf-8")

    # perform request
    status, reason, body = apca_http_post(conn, url, postData)

    # close connection
    conn.close()

    if (status == 200):
        return format_pem_cert_header_trailer(body)

    raise Exception(str(status), body)


def apca_renew_cert(crtreq, uri, machine, local_ip=None):
    """ Request APCA to renew an existing certificate by a CSR """
    print "RENEWING CERTIFICATE"
    # send to a web service to sign the certificate
    conn = HTTPSClient(uri,
                       timeout=5,
                       local_ip=local_ip,
                       keyfile=cert_dir + "/" + machine + ".key",
                       certfile=cert_dir + "/" + machine + ".crt")

    url = apca_renew_cert_ep + apca_get_querystring(machine)

    # removing pre and post strings from certificate request
    crtreq = remove_pem_req_header_trailer(crtreq)
    postData = unicode(crtreq, "utf-8")

    # perform request
    status, reason, body = apca_http_post(conn, url, postData)

    # close connection
    conn.close()

    if (status == 200):
        return format_pem_cert_header_trailer(body)

    raise Exception(str(status), body)


def store_ap_pki_pkcs(data):
    """ This function stores in local file the AP PKI certificates PKCS12 """
    allpemContent = ""
    p12file = cdir_path("ap_pki_store.p12")
    with open(p12file, "w") as f:
        f.write(data)

    try:
        from OpenSSL import crypto
        pkcs12 = crypto.load_pkcs12(data)
        certs = pkcs12.get_ca_certificates()
        for cert in certs:
            cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
            cert_filename = cert.get_subject().commonName.replace(" ", "_") + ".crt"
            with open(cdir_path(cert_filename), "w") as f:
                f.write(cert_pem)
                allpemContent += cert_pem
    except:
        ossl_args = ["openssl", "pkcs12", "-in", p12file, "-passin", "pass:"]
        err = ""
        try:
            p = subprocess.Popen(ossl_args,
                                 shell=False,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            out, err = p.communicate()
        except:
            print "Failed to execute '%s' command. Error:%s" % (ossl_args, err)
            raise

        content = out.splitlines()

        # iterate through the output lines and extract single certificates
        f = None
        cn_hdr = "subject=/CN="
        curr_cn = None
        for line in content:
            if (line.startswith(cn_hdr)):
                curr_cn = line[len(cn_hdr):].replace(" ", "_")
                continue
            elif (curr_cn is not None and line.startswith(pem_cert_header)):
                f = open(cdir_path(curr_cn + ".crt"), "w")

            if (f is not None):
                f.write(line)
                f.write(os.linesep)
                allpemContent += line + os.linesep

                if (line.startswith(pem_cert_trailer)):
                    f.close()
                    f = curr_cn = None

    outpem = cdir_path("ap_pki_store.pem")
    with open(outpem, "w") as f:
        f.write(allpemContent)

    return True


def apca_export_pki_store(uri, local_ip=None):
    """ Exporting PKI store from APCA """
    print "Exporting PKI store"
    # send to a web service to sign the certificate
    conn = HTTPSClient(uri,
                       timeout=5,
                       local_ip=local_ip)

    # perform request
    status, reason, body = apca_http_get(conn, apca_export_pki_ep)

    # close connection
    conn.close()

    if (status == 200):
        if (store_ap_pki_pkcs(body)):
            print "AP PKI stored correctly"
        else:
            print "Failed to store store_ap_pki_pkcs"
    else:
        print "Failed to retrieve pki store from %s" % (uri)
        raise Exception(str(status), body)


def passphrase_callback(v):
    """ This is a callback that returns a passphrase needed to dump a key to file. """
    keystorepass = "swan"
    return keystorepass


def remove_passprhase(in_key_file, out_key_file, passprhase):
    """ This is so we can remove the passphrase from the file key. On the server the key is not protected. """
    ossl_args = ["openssl",
                 "rsa",
                 "-in",
                 cdir_path(in_key_file),
                 "-out",
                 cdir_path(out_key_file),
                 "-passin",
                 "pass:" + passprhase]
    err = ""
    try:
        p = subprocess.Popen(ossl_args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
    except:
        print "Failed to execute '%s' command. Error:%s" % (ossl_args, err)
        raise Exception(err)

    return True


def do_work(fqdn, local_ip, apcaservice, renew=False, days_to_renewal=12, mf_cert=False, machine_dir=""):
    """ Do the actual work """
    # if this is a renew check we have to perform a renewal first
    loc_crt_file = cdir_path(fqdn + ".crt")
    loc_key_file = cdir_path(fqdn + ".key")

    # we cannot renew if the certificate and private keys don't exist
    if not os.path.exists(loc_crt_file) or not os.path.exists(loc_key_file):
        renew = False

    if renew:
        # load the certificate and checks it is within the expire date
        cert = X509.load_cert(loc_crt_file)
        certExpiryUtcTime = str(cert.get_not_after())
        certExpiryUtcTime = datetime.strptime(certExpiryUtcTime,
                                              "%b %d %H:%M:%S %Y %Z")
        utcExpiryLimit = datetime.utcnow() + timedelta(days=days_to_renewal)
        print "Certificate expiry date: %s\
                Expiry limit:%s" % (certExpiryUtcTime, utcExpiryLimit)
        if (datetime.utcnow() > certExpiryUtcTime):
            print "Certificate has expired. Trying to provision a new one"
            renew = False
        elif (utcExpiryLimit > certExpiryUtcTime):
            print "Certificate needs renewal."
            global renew_mf_cert
            renew_mf_cert = True
        else:
            print "No renewal needed."
            return

    print "* Generate RSA Key"
    keyfile_pemencrypted = fqdn + ".keyenc"
    rsa = generate_rsa_key()
    rsa.save_key(cdir_path(keyfile_pemencrypted),
                 cipher="aes_256_cbc",
                 callback=passphrase_callback)

    print "* Generate Pub/Pri Keys"
    pkey = make_private_key(rsa)

    print "* Generate Certificate Request for", fqdn
    req = make_cert_request(pkey, fqdn)

    print "* Generate Certificate Request PEM"
    crtreq = req.as_pem()

    print "* Perform HTTP Request to APCA"
    # print crtreq
    crtresp = None
    if mf_cert:
        crtresp = apca_new_mf_cert(crtreq, apcaservice, fqdn, local_ip, machine_dir)
    elif (not renew):
        crtresp = apca_new_cert(crtreq, apcaservice, fqdn, local_ip)
    else:
        crtresp = apca_renew_cert(crtreq, apcaservice, fqdn, local_ip)

    print "* Certificate Response received"  # , crtresp
    files_to_write = [loc_crt_file, cdir_path("machine.crt")]
    for filename in files_to_write:
        with open(filename, "w") as crtfile:
            crtfile.write(crtresp)

    # Validate certificate can be loaded
    cert = X509.load_cert(loc_crt_file)

    # Save the key unencrypted (works only if openssl is available)
    keyfile_pem_ar = [fqdn + ".key", cdir_path("machine.key")]
    for keyfile_pem in keyfile_pem_ar:
        res = remove_passprhase(keyfile_pemencrypted,
                                keyfile_pem,
                                passphrase_callback(None))
        if (not res):
            rsa.save_key(cdir_path(keyfile_pem),
                         cipher="aes_256_cbc",
                         callback=passphrase_callback)


def usage():
    print "Usage:\n \
            python apca_client.py -n <devicename> [options...]\n\n\
            -n  -- Name of the device requesting the certificate\n\
            -i  -- Local interface IP address to bound to (optional, default=any)\n\
            -s  -- APCA service address (optional, default=APCA-backend.Autopilot.Co3c.ap.gbl:443) \n\
            -d  -- If specified will run as a daemon (optional, default=false)\n\
            -r  -- If specified will try to renew only (optional, default=false)\n\
            -t  -- Specifies the seconds to sleep between cycles when running as deamon (optional, default=600)\n\
            -w  -- Specifies the number of days of early expiration (optional, default=12)\n\
            -e  -- Perform AP PKI CA certificates retrieval. This is normally used for bootstrapping\n\
            -m  -- Perform AP Machine Function certificates retrieval\n\
            -f  -- Specifies the directory where certificates and keys will be read or written (optional, default=.)\n\
           "


if __name__ == "__main__":
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   "hn:i:s:rdt:w:ef:m:",
                                   ["help",
                                    "fqdn=",
                                    "localip=",
                                    "apcaservice=",
                                    "renew",
                                    "deamon",
                                    "sleepsecs=",
                                    "daystorenew=",
                                    "certdir=",
                                    "machinecertdir="])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    fqdn = None
    local_ip = None
    renew = False
    deamon = False
    export_store = False
    sleep_secs = 600
    days_to_renewal = 14
    machine_cert_dir = None
    renew_mf_cert = False
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
            sys.exit()
        elif opt in ("-n", "--fqdn"):
            fqdn = arg
        elif opt in ("-i"):
            local_ip = arg
        elif opt in ("-s"):
            apca_url = arg
        elif opt in ("-r"):
            renew = True
        elif opt in ("-d"):
            deamon = True
        elif opt in ("-t"):
            sleep_secs = int(arg)
        elif opt in ("-w"):
            days_to_renewal = int(arg)
        elif opt in ("-e"):
            export_store = True
        elif opt in ("-m"):
            machine_cert_dir = arg
            renew_mf_cert = True
        elif opt in ("-f"):
            cert_dir = arg
            if not os.path.exists(cert_dir):
                os.makedirs(cert_dir)

    if fqdn is None:
        usage()
        sys.exit(2)

    if export_store:
        apca_export_pki_store(apca_url, local_ip)
    elif renew_mf_cert:
        print "Request a Machine Function Cert"
        do_work(fqdn, local_ip, apca_url, renew, days_to_renewal, renew_mf_cert, machine_cert_dir)
    elif renew:
        # See if we have to renew machine cert
        do_work(fqdn, local_ip, apca_url, True, days_to_renewal, renew_mf_cert)

        # Do we also have to renew Machine Function Cert?
        if renew_mf_cert:
            machine_cert_dir = cert_dir
            cert_dir = machine_cert_dir + "/../machine_function"
            print "Renewing Machine Function certs into " + cert_dir
            if not os.path.exists(cert_dir):
                os.makedirs(cert_dir)
            do_work(fqdn, local_ip, apca_url, False, days_to_renewal, True, machine_cert_dir)

            # Lets get the intermediate APCA certs too
            cert_dir = machine_cert_dir + "/../intermediate"
            if not os.path.exists(cert_dir):
                os.makedirs(cert_dir)

            print "Renewing APCA Intermediate certs into " + cert_dir
            apca_export_pki_store(apca_url, local_ip)
    else:
        if deamon:
            print "Running as a deamon every %d seconds" % (sleep_secs)
            while (True):

                try:
                    do_work(fqdn, local_ip, apca_url, True, days_to_renewal)
                except Exception as exc:
                    print "Exception raised: ", str(exc)

                print "Waiting %d seconds..." % (sleep_secs)
                time.sleep(sleep_secs)
        else:
            do_work(fqdn, local_ip, apca_url, renew, days_to_renewal, renew_mf_cert)
