#!/usr/bin/env python3

# Watchdog used to check Hard Drives and report on their status

import sys
import os
import socket
import subprocess
import io
import itertools
import re
import platform
from collections import OrderedDict


class WatchAPHDD(object):
    def __init__(self, watchdog_csv_dir: str, watchdog_tmp_file: str, watchdog_csv_file: str, test_status=False):

        # If we are in unit test mode we deal with certain situations
        # which we can't replicate with generic data
        self.test_status = test_status

        # machinelocalwatchdog folder and files to use for DM report
        self.watchdog_csv_dir = watchdog_csv_dir
        self.watchdog_tmp_file = watchdog_tmp_file
        self.watchdog_csv_file = watchdog_csv_file

        # Get hostname
        self.hostname = socket.gethostname()

        # Store raw devices
        self.raw_devices = []

        # Store processed results
        self.device_results = {}

    # Use lsblk to determine what block devices are in the system
    def discover_block_devices(self) -> bool:
        if self.test_status:
            # If this is a unit test pass in 6 regular device names
            result = 'sda  disk\nsdb  disk\nsdc  disk\nsdd  disk\nsde  disk\nsdf  disk\nsr0  rom\n'
        else:
            command = "lsblk -S -n -o name,type"

            # print ("\nwatchaphdd:discover_block_devices Command: " + command)
            try:
                result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
            except subprocess.TimeoutExpired:
                return False
            except subprocess.CalledProcessError:
                return False

        raw_devices_list = result.strip().split("\n")

        raw_devices = {}
        for device_data in raw_devices_list:
            raw_devices[device_data.strip().split()[0]] = device_data.strip().split()[1]

        # Make sure devices aren't all null
        device_found = False
        for device, type in raw_devices.items():
            if device:
                device_found = True

        if not device_found:
            return False

        # Make sure devices exclude anything that's not "disk" type
        for device in list(raw_devices):
            device_type = raw_devices[device]
            if device_type != "disk":
                del raw_devices[device]

        # Make sure more than 1 device is found
        if len(raw_devices) < 2:
            return False
        else:
            self.raw_devices = raw_devices

        return True

    # Print list of raw block devices for output
    def print_raw_devices(self):
        try:
            print("\nFound Raw devices:")
            for device in list(filter(None, self.raw_devices)):
                print("\t" + device)
        except Exception as e:
            print("watchaphdd.print_device_results: Exception: " + str(e))
            return False

        return True

    # Check to see if the device is rotational or not
    def get_block_device_type(self) -> bool:
        my_iterator = itertools.cycle(range(2))

        for device in list(filter(None, self.raw_devices)):
            # Determine whether the device is an SSD or an HDD by probing the rotational parameter
            command = "cat /sys/block/" + device + "/queue/rotational"
            # print("\nwatchaphdd:get_block_device_type Getting device type information on /dev/" + device)
            # print("watchaphdd:get_block_device_type " + command)

            if self.test_status:
                # If this is a unit test we alternate block devices and make every other one
                # either 0 or 1, 0 is non-rotational, 1 is rotational, this gives us 3 non-rotational and 3
                # rotational to test
                # We use an iterator object to cycle between 0 and 1
                value = next(my_iterator)
                result = str(value) + '\n'
            else:
                try:
                    result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
                except subprocess.TimeoutExpired:
                    result = ""
                except subprocess.CalledProcessError:
                    result = ""

            device_type_value = result.split("\n")[0]

            if not device_type_value:
                print("\nwatchaphdd:get_block_device_type Unable to determine HDD vs SSD for device: " + device)
                return False

            if (device_type_value == "1"):
                device_type = "HDD"
                # print("watchaphdd:get_block_device_type Found HDD rotational device")
            else:
                device_type = "SSD"
                # print("watchaphdd:get_block_device_type Found SSD device")

            self.device_results[device] = {"type": device_type}

        return True

    # Get the name of the device Bus ID using lsscsi
    def get_lsscsi_device_bus(self) -> bool:
        command = "lsscsi --verbose | egrep " + r'"^\["' + r" | awk '{ print $1 " + '" "' + r" $NF }'"
        # print("\nwatchaphdd:get_lsscsi_device_bus Running: " + command + " to get device Bus IDs")

        if self.test_status:
            result = '[0:0:0:0] /dev/sda\n[1:0:0:0] /dev/sdb\n[2:0:0:0] /dev/sdc\n[3:0:0:0] /dev/sdd\n[4:0:0:0] /dev/sde\n[5:0:0:0] /dev/sdf\n[6:0:0:0] /dev/sr0\n'
        else:
            try:
                result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
            except subprocess.TimeoutExpired:
                result = ""
            except subprocess.CalledProcessError:
                result = ""

        device_list = result.split("\n")

        for device in list(filter(None, device_list)):

            # Format is this: [0:0:0:0]
            raw_bus = device.split(" ")[0]

            # ['0', '0', '0', '0']
            # ['1', '0', '0', '0']
            # ['2', '0', '0', '0']
            # ['3', '0', '0', '0']
            # ['4', '0', '0', '0']
            # ['5', '0', '0', '0']
            # ['6', '0', '0', '0']
            bus_split_values = raw_bus.translate({ord(i): None for i in '[]'}).split(":")

            # 1:0
            bus = ':'.join(bus_split_values[0:2])

            pattern = re.compile('[0-9]:[0-9]')
            match = pattern.search(bus)

            if not bus or not match:
                return False

            # /dev/sda
            raw_device = device.split(" ")[1]

            # ['', 'dev', 'sda']
            device = raw_device.split("/")[2]

            if device.strip() in self.device_results:
                for key, value in self.device_results.items():
                    if key.strip() == device.strip():
                        self.device_results[device]["bus"] = bus

        return True

    # use smartctl to get device health and device detailed information like firmware version, etc.
    def get_smartctl_info(self) -> bool:
        # print("\nwatchaphdd:get_smartctl_info Getting smartctl device information and health")

        for device in list(filter(None, self.raw_devices)):
            device_model = ""
            device_serial = ""
            device_firmware = ""
            result_smart = ""
            smart_error = False

            if self.test_status:
                result_smart = 'SMART overall-health self-assessment test result: PASSED\n'
            else:
                command = "smartctl -H /dev/" + device + " | grep PASSED"
                try:
                    result_smart = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
                except subprocess.TimeoutExpired:
                    smart_error = True
                except subprocess.CalledProcessError:
                    smart_error = True

            # We only do a complete property check on the hard drive if it passes SMART health test
            # since we don't want to lock up the machine if the drive is dying
            if not smart_error:
                result_details = ""

                if self.test_status:
                    result_details = 'Device Model:     SK hynix SH921 2.5 7MM 480GB\nSerial Number:    ' \
                                     '1IE7T000110507B0P\nFirmware Version: 10011L00\n'
                else:
                    command = "smartctl -i /dev/" + device + " | egrep 'Model|Serial|Firmware'"
                    try:
                        result_details = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
                    except subprocess.TimeoutExpired:
                        smart_error = True
                    except subprocess.CalledProcessError:
                        smart_error = True

            # If the device is in error state we mark all properties as being in error because we do not want to parse
            # details on a bad drive as it may lock up the system
            if smart_error or not result_smart:
                device_model = "Error"
                device_serial = "Error"
                device_firmware = "Error"

                if device.strip() in self.device_results:
                    for key, value in self.device_results.items():
                        if key.strip() == device.strip():
                            # Switch to Fatal once we're sure no other weird device names show up
                            # self.device_results[device]["status"] = "ErrorFatal"
                            self.device_results[device]["status"] = "Warning"
                            self.device_results[device]["model"] = device_model
                            self.device_results[device]["serial"] = device_serial
                            self.device_results[device]["firmware"] = device_firmware
            else:
                if not result_details:
                    device_model = "Error"
                    device_serial = "Error"
                    device_firmware = "Error"

                else:
                    # ['Device Model:     SK hynix SH921 2.5 7MM 480GB', 'Serial Number:    1IE7T000110507B0P', Firmware Version: 10011L00', '']
                    raw_device_details = result_details.split("\n")

                    for details in list(filter(None, raw_device_details)):
                        pattern_model = re.compile('Model')
                        pattern_serial = re.compile('Serial')
                        pattern_firmware = re.compile('Firmware')
                        match_model = pattern_model.search(details)
                        match_serial = pattern_serial.search(details)
                        match_firmware = pattern_firmware.search(details)

                        if match_model:
                            device_model = details.split(":")[1].strip()

                        if match_serial:
                            device_serial = details.split(":")[1].strip()

                        if match_firmware:
                            device_firmware = details.split(":")[1].strip()

                        if not match_model and not match_serial and not match_firmware:
                            device_model = "Error"
                            device_serial = "Error"
                            device_firmware = "Error"

                result_smart_stripped = result_smart.strip()
                pattern_smart_passed = re.compile('PASSED')
                match_smart_passed = pattern_smart_passed.search(result_smart_stripped)

                if device.strip() in self.device_results:
                    for key, value in self.device_results.items():
                        if key.strip() == device.strip():
                            if not match_smart_passed:
                                # Switch to Fatal once we're sure no other weird device names show up
                                # self.device_results[device]["status"] = "ErrorFatal"
                                self.device_results[device]["status"] = "Warning"
                            else:
                                self.device_results[device]["status"] = "Ok"

                            self.device_results[device]["model"] = device_model
                            self.device_results[device]["serial"] = device_serial
                            self.device_results[device]["firmware"] = device_firmware
                else:
                    return False

        return True

    # Sort all devices in an ordered dictionary and print out the results
    def print_device_results(self):
        try:
            print("\nDevice Results:")
            print(OrderedDict(sorted(self.device_results.items(), key=lambda x: x[1]['bus'])).items())
        except Exception as e:
            print("watchaphdd.print_device_results: Exception: " + str(e))
            return False

        return True

    # Create DM report out of the information returned
    def build_dm_report(self):
        self.dm_output = io.StringIO()

        # CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
        for device, value in OrderedDict(sorted(self.device_results.items(), key=lambda x: x[1]['bus'])).items():
            line = [self.hostname,
                    self.device_results[device]["status"], "smartTst:" +
                    self.device_results[device]["bus"],
                    self.device_results[device]["type"] + " disk " +
                    self.device_results[device]["bus"] + " model '" +
                    self.device_results[device]["model"] + "' serial '" +
                    self.device_results[device]["serial"] + "' firmware '" +
                    self.device_results[device]["firmware"] + "' is " +
                    self.device_results[device]["status"]]

            try:
                self.dm_output.write(",".join(line) + '\r\n')
            except Exception as e:
                print("watchaphdd.build_dm_report: Exception: " + str(e))
                return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
            if not len(results) == 4:
                print("watchaphdd.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.watchdog_csv_dir):
            print("watchaphdd.write_dm_report_to_disk: Directory: " + self.watchdog_csv_dir + " not found")
            return False
        else:
            try:
                csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

                with open(csv_file_path, "w") as text_file:
                    text_file.write(self.dm_output.getvalue())

            except Exception as e:
                print("watchaphdd.write_dm_report_to_disk: Exception: " + str(e))
                return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'r') as content_file:
                content = content_file.read()

                content = content.split("\n")
                dm_output = self.dm_output.getvalue().split("\r\n")

                if list(filter(None, content)) == list(filter(None, dm_output)):
                    return True
                else:
                    print("watchaphdd.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                    return False

        else:
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_csv_file)

        # print("\nRenaming: " + tmp_file_path + " to: " + csv_file_path)

        if platform.system() == "Windows" and self.test_status:
            return True

        try:
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            print("watchaphdd.atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("watchaphdd.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Execute all the methods in order
    def main(self):
        if not self.discover_block_devices():
            print("watchaphdd.main: Could not discover block devices (main.discover_block_devices)")
            sys.exit(1)

        if not self.print_raw_devices():
            print("watchaphdd.main: Could not print raw devices (main.print_raw_devices)")
            sys.exit(1)

        if not self.get_block_device_type():
            print("watchaphdd.main: Could not get block device types (main.get_block_device_type)")
            sys.exit(1)

        if not self.get_lsscsi_device_bus():
            print("watchaphdd.main: Could not get lsscsi information (main.get_lsscsi_device_bus)")
            sys.exit(1)

        # if not self.print_device_results():
        #     print("watchaphdd.main: Could not print device results (main.print_device_results)")
        #     sys.exit(1)

        if not self.get_smartctl_info():
            print("watchaphdd.main: Could not get smartctl information (main.get_smartctl_info)")
            sys.exit(1)

        if not self.print_device_results():
            print("watchaphdd.main: Could not print device results (main.print_device_results)")
            sys.exit(1)

        if not self.build_dm_report():
            print("watchaphdd.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("watchaphdd.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("watchaphdd.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.verify_write_dm_report_to_disk():
            print("watchaphdd.main: Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print("watchaphdd.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        # if not self.print_dm_report():
        #     print("watchaphdd.main: Could not print DM report (main.print_dm_report)")
        #     sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally

if __name__ == '__main__':
    watchdog_csv_dir = "/ap/data/watchdogoutput/csv_sources"
    watchdog_tmp_file = "watchaphdd.tmp"
    watchdog_csv_file = "watchaphdd.csv"

    WatchAPHDD(watchdog_csv_dir, watchdog_tmp_file, watchdog_csv_file).main()
