#!/usr/bin/env python3

# Watchdog used to check version information and report on it

import sys
import os
import socket
import io
import platform
import subprocess

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apcommon import apthelper


class WatchAPVersion(object):
    def __init__(self, watchdog_csv_dir: str, watchdog_tmp_file: str, watchdog_csv_file: str,
                 osversion_file: str="", test_status: bool=False):

        # If we are in unit test mode we deal with certain situations
        # which we can't replicate with generic data
        self.test_status = test_status

        # Path to OS Version file
        self.osversion_file = osversion_file

        # machinelocalwatchdog folder and files to use for DM report
        self.watchdog_csv_dir = watchdog_csv_dir
        self.watchdog_tmp_file = watchdog_tmp_file
        self.watchdog_csv_file = watchdog_csv_file

        # Get hostname
        self.hostname = socket.gethostname()

        # Prepare dict for version result data
        self.version_results = {}

    # Get the OS version by querying the file placed in the image in
    # /ap/data/OsVersion  This file not existing is an error since it should be imaged
    def get_os_version(self) -> bool:
        try:
            if self.test_status:
                self.version_results["osversion"] = "Ubuntu-16.04_20161001"
            else:
                # print("\nwatchapversion.get_os_version Reading OsVersion file: " + self.osversion_file)
                if os.path.isfile(self.osversion_file):
                    with open(self.osversion_file, 'r') as f:
                        first_line = f.readline()
                        f.close()

                        # The only supported OS right now is Ubuntu so we match that to make sure it's valid
                        if first_line.strip().startswith("Ubuntu"):
                            # print("watchapversion.get_os_version OsVersion matches Ubuntu")
                            self.version_results["osversion"] = first_line.strip()
                        else:
                            print("watchapversion.get_os_version OsVersion does not match Ubuntu")
                            return False
                else:
                    print("watchapversion.get_os_version OsVersion file does not exist")
                    return False

        except Exception as e:
            print("\nwatchapversion: Exception: " + str(e))
            return False

        return True

    # Use apthelper.py library to query shared services version and store it
    def get_shared_services_version(self) -> bool:
        if self.test_status:
            self.version_results["services-version"] = "0.1.7-develop.23"
        else:
            try:
                version = apthelper.apt_get_version("aplinux-shared-services")

                self.version_results["services-version"] = version.strip()
            except Exception as e:
                print("\nwatchapversion: Exception: " + str(e))
                return False

        return True

    # Get existing kernel version from the system
    def get_kernel_version(self) -> bool:
        if self.test_status:
            self.version_results["kernel-version"] = "Linux CO4PHY050071133 4.4.0-66-generic #87-Ubuntu SMP Fri Mar 3 15:29:05 UTC 2017 x86_64 x86_64 x86_64 GNU/Linux"
        else:
            command = "uname -a"
            try:
                result_kernel = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)

                self.version_results["kernel-version"] = result_kernel.strip()
            except subprocess.TimeoutExpired:
                return False
            except subprocess.CalledProcessError:
                return False

        return True

    # Print out version results
    def print_version_results(self):
        try:
            print("\nVersions of OS, Kernel and Linux Shared Services:")
            print(self.version_results)

        except Exception as e:
            print("watchapversion.print_versions: Exception: " + str(e))
            return False

        return True

    # Create DM report out of the information returned
    def build_dm_report(self):
        self.dm_output = io.StringIO()

        monitored_items = {}
        monitored_items["osversion"] = {"OSImageVersion": self.version_results["osversion"]}
        monitored_items["services-version"] = {"LinuxSharedServicesVersion": self.version_results["services-version"]}
        monitored_items["kernel-version"] = {"KernelVersion": self.version_results["kernel-version"]}

        status = "Ok"

        for key, value in monitored_items.items():
            if key in self.version_results:
                if not self.version_results[key]:
                    status = "Warning"

            line = ""
            for key2, value2 in monitored_items[key].items():
                line = [self.hostname, status, key2, value2]

            try:
                self.dm_output.write(",".join(line) + '\r\n')
            except Exception as e:
                print("watchapversion.build_dm_report: Exception: " + str(e))
                return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
            if not len(results) == 4:
                print("watchapversion.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.watchdog_csv_dir):
            print("watchapversion.write_dm_report_to_disk: Directory: " + self.watchdog_csv_dir + " not found")
            return False
        else:
            try:
                csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

                with open(csv_file_path, "w") as text_file:
                    text_file.write(self.dm_output.getvalue())

            except Exception as e:
                print("watchapversion.write_dm_report_to_disk: Exception: " + str(e))
                return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'r') as content_file:
                content = content_file.read()

                content = content.split("\n")
                dm_output = self.dm_output.getvalue().split("\r\n")

                if list(filter(None, content)) == list(filter(None, dm_output)):
                    return True
                else:
                    print("watchapversion.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                    return False

        else:
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_csv_file)

        # print("\nRenaming: " + tmp_file_path + " to: " + csv_file_path)

        if platform.system() == "Windows" and self.test_status:
            return True

        try:
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            print("watchapversion.atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("watchapversion.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Execute all the methods in order
    def main(self):
        if not self.get_os_version():
            print("watchapversion.main: Could not get OS Version (main.get_os_version)")
            sys.exit(1)

        if not self.get_shared_services_version():
            print("watchapversion.main: Could not get shared services version (main.get_shared_services_version)")
            sys.exit(1)

        if not self.get_kernel_version():
            print("watchapversion.main: Could not get kernel version (main.get_kernel_version)")
            sys.exit(1)

        # if not self.print_version_results():
        #     print("watchapversion.main: Could not print version results (main.print_version_results)")
        #     sys.exit(1)

        if not self.build_dm_report():
            print("watchapversion.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("watchapversion.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("watchapversion.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.verify_write_dm_report_to_disk():
            print("watchapversion.main: Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print("watchapversion.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        # if not self.print_dm_report():
        #     print("watchapversion.main: Could not print DM report (main.print_dm_report)")
        #     sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally

if __name__ == '__main__':
    watchdog_csv_dir = "/ap/data/watchdogoutput/csv_sources"
    watchdog_tmp_file = "watchapversion.tmp"
    watchdog_csv_file = "watchapversion.csv"
    osversion_file = "/ap/data/OsVersion"

    WatchAPVersion(watchdog_csv_dir, watchdog_tmp_file, watchdog_csv_file, osversion_file).main()
