#!/usr/bin/env python3

# Watchdog used to check health of AzSecPack (e.g. connection to the mdsd socket)

import sys
import os
import io
import socket
from systemd import journal
from datetime import datetime
from datetime import timedelta
from dateutil import tz


class WatchAzSecPack(object):
    def __init__(self,
                 period_in_minutes: int=30,
                 report_directory: str="/ap/data/watchdogoutput/csv_sources"):
        self.period_in_minutes = period_in_minutes
        self.report_directory = report_directory
        self.hostname = ""

        self.syslog_id = "azsecd"
        self.watchdog_name = "watchazsecpack"
        self.watchdog_tmp_file = self.watchdog_name + ".tmp"
        self.watchdog_csv_file = self.watchdog_name + ".csv"

        # Report property values
        self.recent_errors = []
        self.recent_error_status = "Ok"

    def log_entries(self):
        j = journal.Reader()
        j.log_level(journal.LOG_INFO)
        start_period = datetime.now() - timedelta(minutes=self.period_in_minutes)
        j.seek_realtime(start_period)

        for entry in j:
            if 'MESSAGE' not in entry:
                continue
            if 'SYSLOG_IDENTIFIER' not in entry:
                continue
            if entry['SYSLOG_IDENTIFIER'] == self.syslog_id:
                yield entry['MESSAGE']

    def summarize_log_entries(self):
        self.recent_errors = []
        self.recent_error_status = "Ok"

        # we'll key off of the message to prevent message duplication
        # journal returns older entries first so newer entries will overwrite older ones
        summarized_entries = {}
        try:
            for entry in self.log_entries():
                first_space = entry.find(" ")
                if first_space < 0:
                    continue
                second_space = entry.find(" ", first_space + 1)
                if second_space < 0:
                    continue

                timestamp = entry[:second_space]
                message = entry[second_space + 1:]

                try:
                    dt = datetime.strptime(timestamp, "%Y/%m/%d %H:%M:%S")
                except Exception:
                    raise Exception("expected timestamp in '%Y/%m/%d %H:%M:%S' format")

                # convert local time to utc
                utc = dt.replace(tzinfo=tz.tzlocal()).astimezone(tz.tzutc())
                summarized_entries[message] = utc.strftime("%Y-%m-%d %H:%M:%S")

                # adjust status level to the highest severity seen in the log entries
                if message.startswith("ERR") and self.recent_error_status != "Warning":  # TODO should be "Error" but... you know...
                    self.recent_error_status = "Warning"  # TODO should be "Error" but... you know...
                elif message.startswith("WARN") and self.recent_error_status == "Ok":
                    self.recent_error_status = "Warning"
        except Exception as e:
            raise Exception("WatchAzSecPack.summarize_log_entries: " + str(e))

        # we don't need the dictionary anymore so we'll convert it to a list of strings
        for message, timestamp in summarized_entries.items():
            self.recent_errors.append("{0}: {1}".format(timestamp, message))

    # Create DM report out of the information returned
    def build_dm_report(self):
        self.dm_output = io.StringIO()
        if not self.hostname:
            try:
                self.hostname = socket.gethostname()
            except Exception as e:
                print("build_dm_report: Exception: " + str(e))
                return False

        recent = ";".join(str(x) for x in self.recent_errors)
        line = [self.hostname, self.recent_error_status, "AzSecPackRecentErrors", recent]
        try:
            self.dm_output.write(",".join(line) + '\r\n')
        except Exception as e:
            print("build_dm_report: Exception: " + str(e))
            return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
            if not len(results) == 4:
                print("verify_dm_report: Number of expected columns is incorrect")
                return False

            if results[1] not in {"Ok", "Warning", "Error", "Information"}:
                print("verify_dm_report: Invalid status")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.report_directory):
            print("write_dm_report_to_disk: Directory: " + self.report_directory + " not found")
            return False

        tmp_file_path = os.path.join(self.report_directory, self.watchdog_tmp_file)

        try:
            with open(tmp_file_path, "w") as text_file:
                text_file.write(self.dm_output.getvalue())
        except Exception as e:
            print("write_dm_report_to_disk: Exception: " + str(e))
            return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.report_directory, self.watchdog_tmp_file)

        if not os.path.exists(csv_file_path):
            return False

        dm_output = self.dm_output.getvalue().split("\r\n")
        with open(csv_file_path, 'r') as content_file:
            content = content_file.read().split("\n")

        if list(filter(None, content)) != list(filter(None, dm_output)):
            print("verify_write_dm_report_to_disk: csv content does not match dm_output content")
            return False

        return True

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.report_directory, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.report_directory, self.watchdog_csv_file)

        try:
            os.rename(tmp_file_path, csv_file_path)
        except Exception as e:
            print("atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):
        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("print_dm_report: Exception: " + str(e))
            return False

        return True

    # Execute all the methods in order
    def main(self):
        error_code = 0  # EX_OK = everything executed normally

        try:
            self.summarize_log_entries()
        except Exception as e:
            print("WatchAzSecPack.main: {0}".format(e))
            error_code = 1

        if not self.build_dm_report():
            print("WatchAzSecPack.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("WatchAzSecPack.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("WatchAzSecPack.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print("WatchAzSecPack.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        sys.exit(error_code)

if __name__ == '__main__':
    WatchAzSecPack().main()
